import Data.IntMap.Strict as Map
import Data.Vector as Vector
import Data.List.Split(splitOn)
import Data.List as List
import Control.Parallel
import Control.Parallel.Strategies
import Control.Monad.Par(runPar, spawnP, get)
import System.Exit(die)
import System.Environment(getArgs, getProgName)

getVertices :: IntMap b -> [Key]
getVertices m = Prelude.map (\(i, _) -> i) (Map.toList m)

getVerticesVec :: IntMap b -> Vector Key
getVerticesVec m = Vector.fromList $ getVertices m

getVerticesMap :: IntMap b -> IntMap Key
getVerticesMap m = Map.fromList $ Prelude.map (\(i, _) -> (i,i)) (Map.toList m)

weight :: IntMap (IntMap b) -> Key -> Key -> Maybe b
weight graph i j = do
    rowMap <- Map.lookup i graph
    Map.lookup j rowMap

fromJust :: Num p => Maybe p -> p
fromJust (Just a) = a
fromJust Nothing = 0

myMinList2 :: (Ord a, Num a, Num t, Eq t) =>
                    t -> [(Maybe a, b)] -> Maybe (a, b)
myMinList2 depth xs = myMinList2Helper depth (Vector.fromList xs)
    where
        myMinList2Helper d vxs
            | Vector.length vxs == 0 = Nothing
            | Vector.length vxs == 1 && (fst $ Vector.head vxs) == Nothing = Nothing
            | Vector.length vxs == 1 && (fst $ Vector.head vxs) /= Nothing = Just (fromJust (fst $ Vector.head vxs), (snd $ Vector.head vxs))
            | d == 0 = myMin (myMinList2Helper 0 $ Vector.fromList [Vector.head vxs]) (myMinList2Helper 0 $ Vector.tail vxs)
            | otherwise = par leftMin (myMin leftMin rightMin)
                where
                    lxs = Vector.length vxs
                    splittedVecs = Vector.splitAt (div lxs 2) vxs
                    leftMin = myMinList2Helper ((-) d 1) (fst splittedVecs)
                    rightMin = myMinList2Helper ((-) d 1) (snd splittedVecs)

myMinList1 :: (Ord a, Eq t, Num t) =>
                    t -> [(Maybe a, b)] -> Maybe (a, b)
myMinList1 _ [] = Nothing
myMinList1 _ [(Nothing, _)] = Nothing
myMinList1 _ [(Just x, i)] = Just (x, i)
myMinList1 0 (x:xs) = myMin (myMinList1 (0 :: Int) [x]) (myMinList1 (0 :: Int) xs)
myMinList1 d xs = leftMin `par` rightMin `pseq` (myMin leftMin rightMin)
    where
        lxs = List.length xs
        leftMin = myMinList1 ((-) d 1) (List.take (div lxs 2) xs)
        rightMin = myMinList1 ((-) d 1) (List.drop (div lxs 2) xs)

myMinList0 :: Ord a => [(Maybe a, b)] -> Maybe (a, b)
myMinList0 [] = Nothing
myMinList0 [(Nothing, _)] = Nothing
myMinList0 [(Just x, i)] = Just (x, i)
myMinList0 (x:xs) = myMin (myMinList0  [x]) (myMinList0 xs)

myMin :: Ord a => Maybe (a, b) -> Maybe (a, b) -> Maybe (a, b)
myMin Nothing Nothing = Nothing
myMin Nothing (Just (x, i)) = Just (x, i)
myMin (Just (x, i)) Nothing = Just (x, i)
myMin (Just (x, i)) (Just (y, j))
    | x <= y = Just (x, i)
    | otherwise = Just (y, j)

bellmanFordParV1 :: IntMap (IntMap Int) -> IntMap Key
bellmanFordParV1 graphMap = Map.mapWithKey getParent (bellmanFordHelper (getVertices graphMap))
    where
        bellmanFordHelper vertices = List.foldl' iterRelaxRound (initDistParentMap 0) [1..((List.length vertices) - 1)]
            where
                initDistParentMap i = Map.insert i (0, i) (Map.fromList [])
                iterRelaxRound dpmap1 _ = runPar $ do
                    m <- Map.traverseWithKey (\_ v -> spawnP (relaxAll v)) (getVerticesMap graphMap)
                    traverse get m
                    where
                        relaxAll target = fromJust2 (myMin (Map.lookup target dpmap1) (myMinList0 relaxResult))
                            where
                                relaxResult = (Prelude.map relaxTarget vertices)
                                relaxTarget source =
                                    case (prevDist, edgeWeight) of
                                    (Nothing, Nothing) -> (Nothing, source)
                                    (Nothing, Just _) -> (Nothing, source)
                                    (Just _, Nothing) -> (Nothing, source)
                                    (Just pd, Just ew) -> (Just ((+) pd ew), source)
                                    where
                                        prevDist
                                            | a == Nothing = Nothing
                                            | otherwise = Just (fst $ fromJust2 $ a)
                                            where
                                                a = Map.lookup source dpmap1
                                        edgeWeight = weight graphMap source target
                                fromJust2 (Just a) = a
                                fromJust2 Nothing = (0,0)
        getParent _ (_,i) = i

bellmanFordParV2 :: IntMap (IntMap Int) -> IntMap Key
bellmanFordParV2 graphMap = Map.mapWithKey getParent (bellmanFordHelper (getVertices graphMap))
    where
        bellmanFordHelper vertices = List.foldl' iterRelaxRound (initDistParentMap 0) [1..((List.length vertices) - 1)]
            where
                initDistParentMap i = Map.insert i (0, i) (Map.fromList [])
                iterRelaxRound dpmap1 _ = runPar $ do
                    m <- Map.traverseWithKey (\_ v -> spawnP (relaxAll v)) (getVerticesMap graphMap)
                    traverse get m
                    where
                        relaxAll target = fromJust2 (myMin (Map.lookup target dpmap1) (myMinList0 relaxResult))
                            where
                                relaxResult = (parMap rdeepseq relaxTarget vertices) :: [(Maybe Int, Int)]
                                relaxTarget source =
                                    case (prevDist, edgeWeight) of
                                    (Nothing, Nothing) -> (Nothing, source)
                                    (Nothing, Just _) -> (Nothing, source)
                                    (Just _, Nothing) -> (Nothing, source)
                                    (Just pd, Just ew) -> (Just ((+) pd ew), source)
                                    where
                                        prevDist
                                            | a == Nothing = Nothing
                                            | otherwise = Just (fst $ fromJust2 $ a)
                                            where
                                                a = Map.lookup source dpmap1
                                        edgeWeight = weight graphMap source target
                                fromJust2 (Just a) = a
                                fromJust2 Nothing = (0,0)
        getParent _ (_,i) = i

bellmanFordParV3 :: Int -> IntMap (IntMap Int) -> IntMap Key
bellmanFordParV3 depth graphMap = Map.mapWithKey getParent (bellmanFordHelper (getVertices graphMap))
    where
        bellmanFordHelper vertices = List.foldl' iterRelaxRound (initDistParentMap 0) [1..((List.length vertices) - 1)]
            where
                initDistParentMap i = Map.insert i (0, i) (Map.fromList [])
                iterRelaxRound dpmap1 _ = runPar $ do
                    m <- Map.traverseWithKey (\_ v -> spawnP (relaxAll v)) (getVerticesMap graphMap)
                    traverse get m
                    where
                        relaxAll target = fromJust2 (myMin (Map.lookup target dpmap1) (myMinList1 depth relaxResult))
                            where
                                relaxResult = (parMap rdeepseq relaxTarget vertices) :: [(Maybe Int, Int)]
                                relaxTarget source =
                                    case (prevDist, edgeWeight) of
                                    (Nothing, Nothing) -> (Nothing, source)
                                    (Nothing, Just _) -> (Nothing, source)
                                    (Just _, Nothing) -> (Nothing, source)
                                    (Just pd, Just ew) -> (Just ((+) pd ew), source)
                                    where
                                        prevDist
                                            | a == Nothing = Nothing
                                            | otherwise = Just (fst $ fromJust2 $ a)
                                            where
                                                a = Map.lookup source dpmap1
                                        edgeWeight = weight graphMap source target
                                fromJust2 (Just a) = a
                                fromJust2 Nothing = (0,0)
        getParent _ (_,i) = i

bellmanFordParV4 :: Int -> IntMap (IntMap Int) -> IntMap Key
bellmanFordParV4 depth graphMap = Map.mapWithKey getParent (bellmanFordHelper (getVertices graphMap))
    where
        bellmanFordHelper vertices = List.foldl' iterRelaxRound (initDistParentMap 0) [1..((List.length vertices) - 1)]
            where
                initDistParentMap i = Map.insert i (0, i) (Map.fromList [])
                iterRelaxRound dpmap1 _ = runPar $ do
                    m <- Map.traverseWithKey (\_ v -> spawnP (relaxAll v)) (getVerticesMap graphMap)
                    traverse get m
                    where
                        relaxAll target = fromJust2 (myMin (Map.lookup target dpmap1) (myMinList2 depth relaxResult))
                            where
                                relaxResult = (parMap rdeepseq relaxTarget vertices) :: [(Maybe Int, Int)]
                                relaxTarget source =
                                    case (prevDist, edgeWeight) of
                                    (Nothing, Nothing) -> (Nothing, source)
                                    (Nothing, Just _) -> (Nothing, source)
                                    (Just _, Nothing) -> (Nothing, source)
                                    (Just pd, Just ew) -> (Just ((+) pd ew), source)
                                    where
                                        prevDist
                                            | a == Nothing = Nothing
                                            | otherwise = Just (fst $ fromJust2 $ a)
                                            where
                                                a = Map.lookup source dpmap1
                                        edgeWeight = weight graphMap source target
                                fromJust2 (Just a) = a
                                fromJust2 Nothing = (0,0)
        getParent _ (_,i) = i

bellmanFordParV5 :: Int -> IntMap (IntMap Int) -> IntMap Key
bellmanFordParV5 depth graphMap = Map.mapWithKey getParent (bellmanFordHelper (getVertices graphMap))
    where
        bellmanFordHelper vertices = List.foldl' iterRelaxRound (initDistParentMap 0) [1..((List.length vertices) - 1)]
            where
                initDistParentMap i = Map.insert i (0, i) (Map.fromList [])
                iterRelaxRound dpmap1 _ = runPar $ do
                    m <- Map.traverseWithKey (\_ v -> spawnP (relaxAll v)) (getVerticesMap graphMap)
                    traverse get m
                    where
                        relaxAll target = fromJust2 (myMin (Map.lookup target dpmap1) (minRelaxResult depth vertices))
                            where
                                minRelaxResult _ [] = Nothing
                                minRelaxResult _ [x] = transformForMyMin $ relaxTarget x
                                minRelaxResult 0 (x:xs) = myMin (minRelaxResult 0 [x]) (minRelaxResult 0 xs)
                                minRelaxResult d xs = leftMin `par` rightMin `pseq` (myMin leftMin rightMin)
                                    where
                                        lxs = List.length xs
                                        leftMin = minRelaxResult ((-) d 1) (List.take (div lxs 2) xs)
                                        rightMin = minRelaxResult ((-) d 1) (List.drop (div lxs 2) xs)
                                relaxTarget source =
                                    case (prevDist, edgeWeight) of
                                    (Nothing, Nothing) -> (Nothing, source)
                                    (Nothing, Just _) -> (Nothing, source)
                                    (Just _, Nothing) -> (Nothing, source)
                                    (Just pd, Just ew) -> (Just ((+) pd ew), source)
                                    where
                                        prevDist
                                            | a == Nothing = Nothing
                                            | otherwise = Just (fst $ fromJust2 $ a)
                                            where
                                                a = Map.lookup source dpmap1
                                        edgeWeight = weight graphMap source target
                                transformForMyMin (Nothing, _) = Nothing
                                transformForMyMin (Just x, i) = Just (x, i)
                                fromJust2 (Just a) = a
                                fromJust2 Nothing = (0,0)
        getParent _ (_,i) = i

bellmanFordParV6 :: Int -> IntMap (IntMap Int) -> IntMap Key
bellmanFordParV6 depth graphMap = Map.mapWithKey getParent $ bellmanFordHelper (getVertices graphMap) (getVerticesVec graphMap)
    where
        bellmanFordHelper vertices verticesVec = List.foldl' iterRelaxRound (initDistParentMap 0) [1..((List.length vertices) - 1)]
            where
                initDistParentMap i = Map.insert i (0, i) (Map.fromList [])
                iterRelaxRound dpmap1 _ = runPar $ do
                    m <- Map.traverseWithKey (\_ v -> spawnP (relaxAll v)) (getVerticesMap graphMap)
                    traverse get m
                    where
                        relaxAll target = fromJust2 (myMin (Map.lookup target dpmap1) (minRelaxResult depth verticesVec))
                            where
                                minRelaxResult d vxs
                                    | Vector.length vxs == 0 = Nothing
                                    | Vector.length vxs == 1 = transformForMyMin $ relaxTarget $ Vector.head vxs
                                    | d == 0 = myMin (minRelaxResult 0 $ Vector.fromList [Vector.head vxs]) (minRelaxResult 0 $ Vector.tail vxs)
                                    | otherwise = leftMin `par` rightMin `pseq` (myMin leftMin rightMin)
                                        where
                                            lxs = Vector.length vxs
                                            splittedVecs = Vector.splitAt (div lxs 2) vxs
                                            leftMin = minRelaxResult ((-) d 1) (fst splittedVecs)
                                            rightMin = minRelaxResult ((-) d 1) (snd splittedVecs)
                                relaxTarget source =
                                    case (prevDist, edgeWeight) of
                                    (Nothing, Nothing) -> (Nothing, source)
                                    (Nothing, Just _) -> (Nothing, source)
                                    (Just _, Nothing) -> (Nothing, source)
                                    (Just pd, Just ew) -> (Just ((+) pd ew), source)
                                    where
                                        prevDist
                                            | a == Nothing = Nothing
                                            | otherwise = Just (fst $ fromJust2 $ a)
                                            where
                                                a = Map.lookup source dpmap1
                                        edgeWeight = weight graphMap source target
                                transformForMyMin (Nothing, _) = Nothing
                                transformForMyMin (Just x, i) = Just (x, i)
                                fromJust2 (Just a) = a
                                fromJust2 Nothing = (0,0)
        getParent _ (_,i) = i

getCycleFromVertex :: IntMap Key -> [Key]
getCycleFromVertex parentMap = 0 : (f parentMap 0 (Map.insert 0 (1 :: Int) (Map.fromList [])))
    where
        f pMap currentVertex visited
            | parent == Nothing = []
            | Map.lookup (fromJust parent) visited /= Nothing = [fromJust parent]
            | otherwise = (fromJust parent) : f pMap (fromJust parent) (Map.insert (fromJust parent) (1 :: Int) visited)
            where
                parent = Map.lookup currentVertex parentMap

pruneCycle :: Eq a => [a] -> [a]
pruneCycle c
    | c == [] = []
    | otherwise = processCycleHelper c (List.last c)
        where
            processCycleHelper cc st
                | List.head cc == st = List.reverse cc
                | otherwise = processCycleHelper (List.tail cc) st

getCycleWeight :: Num p => [Key] -> IntMap (IntMap p) -> p
getCycleWeight c g
    | (List.length c) < 2 = 0
    | otherwise = (+) (fromJust (weight g (List.head c) (List.head tc))) (getCycleWeight tc g)
        where
            tc = List.tail c

getCycle :: Num a =>
                  [[a]] -> (IntMap (IntMap a) -> IntMap Key) -> ([Key], a)
getCycle testcase parF = cyc
    where
        cyc = ((\c -> (c, getCycleWeight c graphM)) . pruneCycle . getCycleFromVertex) parentMap
        parentMap = parF graphM
        graphM = arrayToMap testcase
        arrayToMap arr = Map.fromList $ Prelude.map f (List.zip [0..] arr)
            where
                f (i, row) = (i, Map.fromList $ List.zip [0..] row)

main :: IO ()
main = do
    args <- getArgs
    case args of
        [version, d, filename] -> do
            contents <- readFile filename
            let f line = ((Prelude.map (\x -> read x :: Int)) . (splitOn ",")) line
            let testCase = (Prelude.map f (lines contents))
            let depth = read d :: Int
            case version of
                "2" -> do
                    print "using v2"
                    print (getCycle testCase bellmanFordParV2)
                "3" -> do
                    print ("using v3 with depth == " List.++ d)
                    print (getCycle testCase (bellmanFordParV3 depth))
                "4" -> do
                    print ("using v4 with depth == " List.++ d)
                    print (getCycle testCase (bellmanFordParV4 depth))
                "5" -> do
                    print ("using v5 with depth == " List.++ d)
                    print (getCycle testCase (bellmanFordParV5 depth))
                "6" -> do
                    print ("using v6 with depth == " List.++ d)
                    print (getCycle testCase (bellmanFordParV6 depth))
                _ -> do
                    print "using v1"
                    print (getCycle testCase bellmanFordParV1)
        _ -> do
            pn <- getProgName
            die $ "Usage: " List.++ pn List.++ " <version> <depth> <filename>"